from subprocess import Popen, PIPE
from datetime import datetime
import socket
import json
import os


def ts2time(ts):
    return datetime.fromtimestamp(ts).strftime('%H:%M:%S')


def raise_on_error(resp):
    if 'error' in resp and resp['error'] != 'none':
        raise Exception(resp['error'])


def generate_key(length):
    key = b''
    with open('/dev/urandom', 'rb') as f:
        while len(key) < length:
            x = f.read(1)
            if x.isalnum():
                key += x
    return str(key, 'ascii')


def get_ip():
    s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
    s.connect(('1.2.3.4', 0))
    return s.getsockname()[0]


def to_credstr(ip, port, publicKey, password, strict=False, **kwargs):
    addr = '%s:%d' % (ip, int(port))
    kwargs['password'] = password
    kwargs['publicKey'] = publicKey
    cred = json.dumps({addr: kwargs})
    return cred if strict else cred[1:-1]


def grep_ns(ns, addr):
    return [x for x in ns if x['addr'] == addr]


def get_from_route(route, key, default):
    return route[0][key] if route else default


def get_version(route):
    return get_from_route(route, 'version', '??')


def get_path(route):
    return get_from_route(route, 'path', 'NO ROUTE TO HOST')


def dns_resolve(addr):
    addr = addr.split(':')
    addr[0] = socket.gethostbyname(addr[0])
    return ':'.join(addr)


def load_conf(conf, bin):
    try:
        with open(conf, 'rb') as f:
            conf = f.read()

        p = Popen([bin, '--cleanconf'], stdin=PIPE, stdout=PIPE)
        conf = p.communicate(conf)[0]
        try:
            conf = str(conf, 'ascii')
        except:
            pass
        return json.loads(conf)
    except ValueError:
        raise Exception('failed to load cjdroute.conf as json')


def load_admin_pw(conf, bin):
    try:
        return os.environ['CJDNS_ADMIN_PW']
    except KeyError:
        conf = load_conf(conf, bin)
        return conf['admin']['password']


def speed(b):
    for unit in ['B', 'Kb', 'Mb', 'Gb', 'Tb']:
        if b < 1024:
            break
        b = round(b / 1024, 2)

    return '%d %s/s' % (b, unit)


def nodeinfo(ip, hub=False, timeout=20):
    import requests

    if hub:
        url = 'http://api.hyperboria.net/v0/node/info.json?ip=%s' % ip
        title = 'hub.hyperboria.net'
    else:
        url = 'http://[%s]/nodeinfo.json' % ip
        title = 'nodeinfo.json'

    j = requests.get(url, timeout=timeout).json
    if not type(j) is dict:
        j = j()
    return j, title



#!/usr/bin/env python
# You may redistribute this program and/or modify it under the terms of
# the GNU General Public License as published by the Free Software Foundation,
# either version 3 of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""cjdns
Allows communication with an instance of cjdns via the admin/RPC port.

Usage:

import cjdns
cjdns = cjdns.connectWithAdminInfo()

"""
from .cjdns import connect, connectWithAdminInfo, PublicToIp6



#!/usr/bin/env python
# You may redistribute this program and/or modify it under the terms of
# the GNU General Public License as published by the Free Software Foundation,
# either version 3 of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""python-cjdns is a library for communicating with the cjdns admin interface"""

from __future__ import print_function

import os
import sys
import socket
import hashlib
import json
import threading
import time
try:
    import queue
except ImportError:
    import Queue as queue
import random
import string
from hashlib import sha512
from .bencode import bencode, bdecode

BUFFER_SIZE = 69632
KEEPALIVE_INTERVAL_SECONDS = 2


class Session(object):
    """Current cjdns admin session"""

    def __init__(self, s):
        self.socket = s
        self.queue = queue.Queue()
        self.messages = {}

    def disconnect(self):
        """Disconnects the socket"""
        self.socket.close()

    def getMessage(self, txid):
        """Retreives the message associatd with txid"""
        return _getMessage(self, txid)

    def functions(self):
        """Prints a list of functions available"""
        print(self._functions)


def _randomString():
    """Random string for message signing"""

    return ''.join(
        random.choice(string.ascii_uppercase + string.digits)
        for x in range(10))


def _callFunc(session, funcName, password, args):
    """Call custom cjdns admin function"""

    txid = _randomString()
    sock = session.socket
    sock.send(bytearray('d1:q6:cookie4:txid10:%se' % txid, 'utf-8'))
    msg = _getMessage(session, txid)
    cookie = msg['cookie']
    txid = _randomString()
    tohash = (password + cookie).encode('utf-8')
    req = {
        'q': funcName,
        'hash': hashlib.sha256(tohash).hexdigest(),
        'cookie': cookie,
        'args': args,
        'txid': txid
    }

    if password:
        req['aq'] = req['q']
        req['q'] = 'auth'
        reqBenc = bencode(req).encode('utf-8')
        req['hash'] = hashlib.sha256(reqBenc).hexdigest()

    reqBenc = bencode(req)
    sock.send(bytearray(reqBenc, 'utf-8'))
    return _getMessage(session, txid)


def _receiverThread(session):
    """Receiving messages from cjdns admin server"""

    timeOfLastSend = time.time()
    timeOfLastRecv = time.time()
    try:
        while True:
            if timeOfLastSend + KEEPALIVE_INTERVAL_SECONDS < time.time():
                if timeOfLastRecv + 10 < time.time():
                    raise Exception("ping timeout")
                session.socket.send(
                    b'd1:q18:Admin_asyncEnabled4:txid8:keepalive')
                timeOfLastSend = time.time()

            try:
                data = session.socket.recv(BUFFER_SIZE)
            except socket.timeout:
                continue

            try:
                benc = bdecode(data)
            except (KeyError, ValueError):
                print("error decoding [" + data + "]")
                continue

            if benc['txid'] == 'keepaliv':
                if benc['asyncEnabled'] == 0:
                    raise Exception("lost session")
                timeOfLastRecv = time.time()
            else:
                # print "putting to queue " + str(benc)
                session.queue.put(benc)

    except KeyboardInterrupt:
        print("interrupted")
        import thread
        thread.interrupt_main()


def _getMessage(session, txid):
    """Getting message associated with txid"""

    while True:
        if txid in session.messages:
            msg = session.messages[txid]
            del session.messages[txid]
            return msg
        else:
            # print "getting from queue"
            try:
                # apparently any timeout at all allows the thread to be
                # stopped but none make it unstoppable with ctrl+c
                nextMessage = session.queue.get(timeout=100)
            except queue.Empty:
                continue
            if 'txid' in nextMessage:
                session.messages[nextMessage['txid']] = nextMessage
                # print "adding message [" + str(next) + "]"
            else:
                print("message with no txid: %s" % nextMessage)


def _functionFabric(func_name, argList, oargList, password):
    """Function fabric for Session class"""

    def functionHandler(self, *args, **kwargs):
        call_args = {}

        for (key, value) in oargList.items():
            call_args[key] = value

        for i, arg in enumerate(argList):
            if i < len(args):
                call_args[arg] = args[i]

        for (key, value) in kwargs.items():
            call_args[key] = value

        return _callFunc(self, func_name, password, call_args)

    functionHandler.__name__ = str(func_name)
    return functionHandler


def connect(ipAddr, port, password):
    """Connect to cjdns admin with this attributes"""

    sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
    sock.connect((ipAddr, port))
    sock.settimeout(2)

    # Make sure it pongs.
    sock.send(b'd1:q4:pinge')
    data = sock.recv(BUFFER_SIZE)
    if not data.endswith(b'1:q4:ponge'):
        raise Exception(
            "Looks like " + ipAddr + ":" + str(port) +
            " is to a non-cjdns socket.")

    # Get the functions and make the object
    page = 0
    availableFunctions = {}
    while True:
        sock.send(bytearray(
            'd1:q24:Admin_availableFunctions4:argsd4:pagei%seee' % page,
            'utf-8'))
        data = sock.recv(BUFFER_SIZE)
        benc = bdecode(data)
        for func in benc['availableFunctions']:
            availableFunctions[func] = benc['availableFunctions'][func]
        if 'more' not in benc:
            break
        page = page+1

    funcArgs = {}
    funcOargs = {}

    for (i, func) in availableFunctions.items():
        items = func.items()

        # grab all the required args first
        # append all the optional args
        rargList = [arg for arg, atts in items if atts['required']]
        argList = rargList + [
            arg for arg, atts in items if not atts['required']]

        # for each optional arg setup a default value with
        # a type which will be ignored by the core.
        oargList = {}
        for (arg, atts) in items:
            if not atts['required']:
                oargList[arg] = (
                    "''" if func[arg]['type'] == 'Int'
                    else "")

        setattr(Session, i, _functionFabric(
            i, argList, oargList, password))

        funcArgs[i] = rargList
        funcOargs[i] = oargList

    session = Session(sock)

    kat = threading.Thread(target=_receiverThread, args=[session])
    kat.setDaemon(True)
    kat.start()

    # Check our password.
    ret = _callFunc(session, "ping", password, {})
    if 'error' in ret:
        raise Exception(
            "Connect failed, incorrect admin password?\n" + str(ret))

    session._functions = ""

    funcOargs_c = {}
    for func in funcOargs:
        funcOargs_c[func] = list([
            key + "=" + str(value) for (key, value) in funcOargs[func].items()
        ])

    for func in availableFunctions:
        session._functions += (
            func + "(" + ', '.join(funcArgs[func] + funcOargs_c[func]) + ")\n")

    # print session.functions
    return session


def connectWithAdminInfo(path=None):
    """Connect to cjdns admin with data from user file"""

    if path is None:
        path = os.path.expanduser('~/.cjdnsadmin')
    try:
        with open(path, 'r') as adminInfo:
            data = json.load(adminInfo)
    except IOError:
        print('~/.cjdnsadmin not found; using default credentials', file=sys.stderr)
        data = {
            'password': 'NONE',
            'addr': '127.0.0.1',
            'port': 11234,
        }

    return connect(data['addr'], data['port'], data['password'])


def Base32_decode(decodeme):
    output = bytearray(len(decodeme))
    numForAscii = [
        99, 99, 99, 99, 99, 99, 99, 99, 99, 99, 99, 99, 99, 99, 99, 99,
        99, 99, 99, 99, 99, 99, 99, 99, 99, 99, 99, 99, 99, 99, 99, 99,
        99, 99, 99, 99, 99, 99, 99, 99, 99, 99, 99, 99, 99, 99, 99, 99,
        0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 99, 99, 99, 99, 99, 99,
        99, 99, 10, 11, 12, 99, 13, 14, 15, 99, 16, 17, 18, 19, 20, 99,
        21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31, 99, 99, 99, 99, 99,
        99, 99, 10, 11, 12, 99, 13, 14, 15, 99, 16, 17, 18, 19, 20, 99,
        21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31, 99, 99, 99, 99, 99,
    ]

    outputIndex = 0
    inputIndex = 0
    nextByte = 0
    bits = 0

    while inputIndex < len(decodeme):
        o = ord(decodeme[inputIndex])
        if o & 0x80:
            raise ValueError
        b = numForAscii[o]
        inputIndex += 1
        if b > 31:
            raise ValueError("bad character " + decodeme[inputIndex])

        nextByte |= (b << bits)
        bits += 5

        if bits >= 8:
            output[outputIndex] = nextByte & 0xff
            outputIndex += 1
            bits -= 8
            nextByte >>= 8

    if bits >= 5 or nextByte:
        raise ValueError("bits is %s and nextByte is %s" % (bits, nextByte))

    return memoryview(output)[0:outputIndex]


def PublicToIp6(pubKey):
    if pubKey[-2:] != ".k":
        raise ValueError("key does not end with .k")
    keyBytes = Base32_decode(pubKey[0:-2])
    hashOne = sha512(keyBytes).digest()
    hashTwo = sha512(hashOne).hexdigest()
    first16 = hashTwo[0:32]
    out = ''
    for i in range(0, 8):
        out += first16[i*4: i*4+4] + ":"
    return out[:-1]



from .cjdns.bencode import bencode, bdecode
from . import cjdns
from hashlib import sha512, sha256
import socket
import os

BUFFER_SIZE = 69632


class Cjdroute(object):
    def __init__(self, ip='127.0.0.1', port=11234, password=''):
        self.password = password

        self.s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        self.s.connect((ip, port))
        self.s.settimeout(7)

        if not self.ping():
            raise Exception('Not a cjdns socket? (%s:%d)' % (ip, port))

    def disconnect(self):
        self.s.close()

    def recv(self):
        res = bdecode(self.s.recv(BUFFER_SIZE))
        if 'error' in res and res['error'] != 'none':
            raise Exception(repr(res))
        if os.getenv('YRD_DEBUG'):
            print(repr(res))  # DEBUG SWITCH
        return res

    def _send(self, **kwargs):
        self.s.send(bencode(kwargs).encode('ascii'))

    def send(self, **kwargs):
        if self.password:
            self._send(q='cookie')
            cookie = self.recv()['cookie']

            kwargs['hash'] = sha256((self.password + cookie).encode('ascii')).hexdigest()
            kwargs['cookie'] = cookie

            kwargs['aq'] = kwargs['q']
            kwargs['q'] = 'auth'
            kwargs['hash'] = sha256(bencode(kwargs).encode('ascii')).hexdigest()

        if os.getenv('YRD_DEBUG'):
            print(repr(kwargs))  # DEBUG SWITCH

        self._send(**kwargs)

    def poll(self, **kwargs):
        if 'args' not in kwargs:
            kwargs['args'] = {}
        kwargs['args']['page'] = 0

        while True:
            self.send(**kwargs)
            resp = self.recv()

            yield resp

            if 'more' not in resp:
                break

            kwargs['args']['page'] += 1

    def ping(self):
        self.send(q='ping')
        resp = self.recv()
        return 'q' in resp and resp['q'] == 'pong'

    def nodeForAddr(self, ip=None):
        q = dict(q='NodeStore_nodeForAddr')
        if ip:
            q['ip'] = ip
        self.send(**q)
        return self.recv()

    def dumpTable(self):
        for page in self.poll(q='NodeStore_dumpTable'):
            for route in page['routingTable']:
                yield route

    def sessionStats(self):
        for page in self.poll(q='SessionManager_getHandles'):
            for handle in page['handles']:
                self.send(q='SessionManager_sessionStats', args={
                    'handle': handle
                })
                yield self.recv()

    def search(self, addr, count=-1):
        self.send(q='SearchRunner_search',
                  args={'ipv6': addr, 'maxRequests': count})

        while True:
            x = self.recv()
            if 'complete' in x:
                break
            yield x

    def genericPing(self, q, path, timeout=5000):
        self.send(q=q, args={'path': path, 'timeout': timeout})
        return self.recv()

    def routerPing(self, *args, **kwargs):
        return self.genericPing('RouterModule_pingNode', *args, **kwargs)

    def switchPing(self, *args, **kwargs):
        return self.genericPing('SwitchPinger_ping', *args, **kwargs)

    def nextHop(self, target, lastNode):
        self.send(q='RouterModule_nextHop',
                  args={'target': target, 'nodeToQuery': lastNode})
        return self.recv()

    def getLink(self, target, num):
        self.send(q='NodeStore_getLink', args={'parent': target,
                                               'linkNum': num})
        return self.recv()

    def addPassword(self, name, password):
        self.send(q='AuthorizedPasswords_add',
                  args={'user': str(name), 'password': str(password)})

    def listPasswords(self):
        self.send(q='AuthorizedPasswords_list')
        return self.recv()

    def removePassword(self, user):
        self.send(q='AuthorizedPasswords_remove', args={'user': user})
        return self.recv()

    def udpBeginConnection(self, addr, pk, password):
        self.send(q='UDPInterface_beginConnection', args={'password': password,
                  'publicKey': pk, 'address': addr})
        return self.recv()

    def peerStats(self):
        for page in self.poll(q='InterfaceController_peerStats'):
            for i, args in page.items():
                if i == 'peers':
                    for peer in args:
                        yield Peer(**peer)


class Peer(object):
    def __init__(self, **kwargs):
        if 'ip' not in kwargs:
            if 'publicKey' in kwargs:
                kwargs['ip'] = pk2ipv6(kwargs['publicKey'])
            elif 'addr' in kwargs:
                kwargs['ip'] = addr2ip(kwargs['addr'])

        if 'path' not in kwargs and 'addr' in kwargs:
            kwargs['path'] = kwargs['addr'][4:23]

        if 'version' not in kwargs and 'addr' in kwargs:
            kwargs['version'] = kwargs['addr'][1:3]

        for x, y in kwargs.items():
            setattr(self, x, y)


def connect(ip='127.0.0.1', port=11234, password=''):
    return Cjdroute(ip, port, password)


# see util/Base32.h
def Base32_decode(input):
    output = bytearray(len(input))
    numForAscii = [
        99, 99, 99, 99, 99, 99, 99, 99, 99, 99, 99, 99, 99, 99, 99, 99,
        99, 99, 99, 99, 99, 99, 99, 99, 99, 99, 99, 99, 99, 99, 99, 99,
        99, 99, 99, 99, 99, 99, 99, 99, 99, 99, 99, 99, 99, 99, 99, 99,
        0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 99, 99, 99, 99, 99, 99,
        99, 99, 10, 11, 12, 99, 13, 14, 15, 99, 16, 17, 18, 19, 20, 99,
        21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31, 99, 99, 99, 99, 99,
        99, 99, 10, 11, 12, 99, 13, 14, 15, 99, 16, 17, 18, 19, 20, 99,
        21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31, 99, 99, 99, 99, 99,
    ]

    outputIndex = 0
    inputIndex = 0
    nextByte = 0
    bits = 0

    while inputIndex < len(input):
        o = ord(input[inputIndex])
        if o & 0x80:
            raise ValueError
        b = numForAscii[o]
        inputIndex += 1
        if b > 31:
            raise ValueError("bad character " + input[inputIndex])

        nextByte |= b << bits
        bits += 5

        if bits >= 8:
            output[outputIndex] = nextByte & 0xff
            outputIndex += 1
            bits -= 8
            nextByte >>= 8

    if bits >= 5 or nextByte:
        raise ValueError("bits is %d and nextByte is %d" % (bits, nextByte))

    return buffer(output, 0, outputIndex)


def addr2ip(addr):
    return pk2ipv6(addr.split('.', 5)[-1])


pk2ipv6 = cjdns.PublicToIp6


def collect_from_address(addr):
    addrs = {}
    parts = len(addr.split('.'))

    if parts == 7:
        addrs['path'] = addr
    elif parts == 2:
        addrs['key'] = addr
    elif parts == 1:
        addrs['ip'] = addr
    else:
        raise ValueError('weird input')

    if 'path' in addrs and 'key' not in addrs:
        addrs['key'] = addrs['path'].split('.', 5)[-1]

    if 'key' in addrs and 'ip' not in addrs:
        addrs['ip'] = pk2ipv6(addrs['key'])

    return addrs



from .xcjdns import addr2ip, collect_from_address

def test_addr2ip():
    addrs = [
        ('v20.0000.0000.0000.0017.c5bu0npp8by4jym96mh0vyy81sn9lhbc01f445nvz64dvdjt98j0.k',
            'fc72:e647:378a:35b6:c343:da8f:db16:3f92'),
        ('v0.0000.0000.0000.0015.1nctdb89gtfrlnu71zyq97n14frl1r4z0ylwzc8vn7kpvrzu4yl0.k',
            'fc00:0000:28a7:1600:168d:4349:4d28:ba73'),
    ]

    for addr, expected in addrs:
        print('trying %r' % addr)
        ip = addr2ip(addr)
        print('got %r' % ip)
        assert ip == expected


def test_collect_from_address():
    addrs = [
        ('v20.0000.0000.0000.0017.c5bu0npp8by4jym96mh0vyy81sn9lhbc01f445nvz64dvdjt98j0.k', {
            'path': 'v20.0000.0000.0000.0017.c5bu0npp8by4jym96mh0vyy81sn9lhbc01f445nvz64dvdjt98j0.k',
            'key': 'c5bu0npp8by4jym96mh0vyy81sn9lhbc01f445nvz64dvdjt98j0.k',
            'ip': 'fc72:e647:378a:35b6:c343:da8f:db16:3f92'}),
        ('v0.0000.0000.0000.0015.1nctdb89gtfrlnu71zyq97n14frl1r4z0ylwzc8vn7kpvrzu4yl0.k', {
            'path': 'v0.0000.0000.0000.0015.1nctdb89gtfrlnu71zyq97n14frl1r4z0ylwzc8vn7kpvrzu4yl0.k',
            'key': '1nctdb89gtfrlnu71zyq97n14frl1r4z0ylwzc8vn7kpvrzu4yl0.k',
            'ip': 'fc00:0000:28a7:1600:168d:4349:4d28:ba73'}),
    ]

    for addr, expected in addrs:
        print('trying %r' % addr)
        stuff = collect_from_address(addr)
        print('got %r' % stuff)
        assert stuff == expected



from .arg import arg, wrap_errors, aliases
from .const import CJDROUTE_CONF, CJDROUTE_BIN
from . import xcjdns as cjdns
from . import cjdns as cj
from . import utils
import itertools
import socket
import time
import os


@arg('-i', '--ip', help='show as ipv6')
@arg('-k', '--key', help='show as key')
@arg('-p', '--path', help='show as path')
@arg('addr', nargs='?', help='address to convert')
@aliases('a')
@wrap_errors([socket.error, IOError, ValueError])
def address(addr, ip=False, key=False, path=False):
    'show cjdroute addresses'

    if addr:
        addrs = cjdns.collect_from_address(addr)

        try:
            if path:
                addr = addrs['path']

            if key:
                addr = addrs['key']

            if ip:
                addr = addrs['ip']
        except KeyError:
            raise ValueError('not enough info')

        return addr
    else:
        c = cjdns.connect()

        res = c.nodeForAddr()['result']
        # TODO: add 'addr' to nodeForAddr
        my_path = 'v%s.%s.%s' % (res['protocolVersion'], res['routeLabel'], res['key'])

        c.disconnect()
        return address(my_path, ip=ip, key=key, path=path)


@arg('ip', help='the cjdns ipv6')
@arg('-c', '--count', metavar='count', help='stop after `count` packets')
@arg('-s', '--switch', help='do a switch ping instead of a router ping')
@wrap_errors([KeyboardInterrupt, socket.error, IOError])
def ping(ip, count=0, switch=False):
    'ping a node'
    admin_pw = utils.load_admin_pw(CJDROUTE_CONF, CJDROUTE_BIN)
    c = cjdns.connect(password=admin_pw)

    ping = c.switchPing if switch else c.routerPing

    for _ in range(count) if count else itertools.repeat(None):
        try:
            resp = ping(ip)
        except Exception as e:
            resp = {'error': e}

        if 'error' in resp:
            yield 'Error: %s' % resp['error']
        elif resp['result'] == 'pong' and switch:
            yield 'Reply from %s %dms' % (resp['path'], resp['ms'])
        elif resp['result'] == 'pong':
            yield 'Reply from %s %dms' % (resp['addr'], resp['ms'])
        elif resp['result'] == 'timeout':
            yield 'Timeout from %s after %dms' % (ip, resp['ms'])

        time.sleep(1)

    c.disconnect()


@arg('-i', '--ip', help='format as ipv6')
@arg('-f', '--follow', help='show new nodes while they\'re discovered')
@aliases('r')
@wrap_errors([KeyboardInterrupt, socket.error])
def route(ip=False, follow=False):
    'access the nodestore'
    c = cjdns.connect()

    known = []
    FMT = '%s %s  v%d %11d %7d'

    while True:
        for node in c.dumpTable():
            if ip:
                if node['ip'] not in known:
                    yield FMT % (node['ip'], node['path'], node['version'],
                                 node['link'], node['time'])
                    known.append(node['ip'])
            else:
                if node['addr'] not in known:
                    yield node['addr']
                    known.append(node['addr'])

        if not follow:
            break

        time.sleep(3)

    c.disconnect()


@arg('-i', '--ip', help='format as ipv6')
@arg('-n', '--neighbours', help='show neighbours peers')
@aliases('n')
@wrap_errors([socket.error, IOError, KeyboardInterrupt])
def neighbours(ip=False, neighbours=False):
    'shows your neighbours'
    c = cjdns.connect()

    if ip:
        STAT_FORMAT = '%s %19s  v%-2s  in %4dkb/s out %4dkb/s  %12s  %d/%d/%d  '
        nodestore = list(c.dumpTable())
    else:
        STAT_FORMAT = '%-78s  in %4dkb/s out %4dkb/s  %12s  %d/%d/%d  '

    for peer in c.peerStats():
        if ip:
            line = STAT_FORMAT % (peer.ip, peer.path, peer.version,
                                  peer.recvKbps, peer.sendKbps, peer.state,
                                  peer.duplicates, peer.lostPackets,
                                  peer.receivedOutOfRange)
        else:
            line = STAT_FORMAT % (peer.addr,
                                  peer.recvKbps, peer.sendKbps, peer.state,
                                  peer.duplicates, peer.lostPackets,
                                  peer.receivedOutOfRange)

        if hasattr(peer, 'user'):
            line += repr(peer.user)

        yield line

        if neighbours:
            links = c.nodeForAddr(peer.addr)['result']['linkCount']
            for i in range(links):
                try:
                    link = c.getLink(peer.ip, i)

                    if link and 'child' in link['result']:
                        child = link['result']['child']
                        if ip:
                            x = cjdns.Peer(addr=child)
                            yield '   %s   %s  v%s' % (x.ip, x.path, x.version)
                        else:
                            yield '   ' + child
                    else:
                        yield '   -'
                except:
                    # TODO remove this
                    pass

    c.disconnect()


@arg('-n', help='number of routes displayed')
@arg('-r', '--routing', help='show routing only')
@wrap_errors([KeyboardInterrupt])
def top(n=25, routing=False):
    'much colors'
    from .top import Session

    s = Session()

    while True:
        print(s.output([] if routing else neighbours(ip=True), route(ip=True), n))
        time.sleep(1)


@arg('-l', '--level', help='filter by log level')
@arg('-f', '--file', help='filter by source file')
@arg('-n', '--line', help='filter by line number')
@arg('-@', dest='about', metavar='pub.k', help='logs related to node')
@wrap_errors([KeyboardInterrupt, IOError])
def mon(level=None, file=None, line=0, about=None):
    'monitor cjdroute'

    admin_pw = utils.load_admin_pw(CJDROUTE_CONF, CJDROUTE_BIN)
    c = cj.connect('127.0.0.1', 11234, admin_pw)

    kwargs = {}

    if level:
        kwargs['level'] = level

    if file:
        kwargs['file'] = file

    if line:
        kwargs['line'] = line

    resp = c.AdminLog_subscribe(**kwargs)
    streamId = resp['streamId']

    if about:
        key = address(about, key=True)
        ip = address(about, ip=True)
        topic = [key, key[:-2], ip]
    else:
        topic = []

    try:
        while True:
            msg = c.getMessage(resp['txid'])
            msg['ftime'] = utils.ts2time(msg['time'])
            if not topic or any([x in msg['message'] for x in topic]):
                yield '{ftime} {level} {file}:{line} {message}'.format(**msg)
    except:
        c.AdminLog_unsubscribe(streamId)
        raise


@wrap_errors([socket.error, IOError])
def sessions():
    'show established sessions'
    admin_pw = utils.load_admin_pw(CJDROUTE_CONF, CJDROUTE_BIN)
    c = cjdns.connect('127.0.0.1', 11234, admin_pw)

    for session in c.sessionStats():
        yield '{addr} {state} {handle} {sendHandle}'.format(**session)


@arg('-c', '--count')
@wrap_errors([socket.error, IOError])
def search(addr, count=-1):
    'search dht for node'
    admin_pw = utils.load_admin_pw(CJDROUTE_CONF, CJDROUTE_BIN)
    c = cjdns.connect('127.0.0.1', 11234, admin_pw)

    count = -1
    for x in c.search(addr, count):
        x['num'] = len(x['nodes'])
        yield '{from} {ms}ms {num} results'.format(**x)
        for y in x['nodes']:
            yield '    %s' % y


@arg('-i', '--ip', help='format as ipv6')
@wrap_errors([IOError])
def uplinks(addr, ip=False):
    'show uplinks of a node'
    admin_pw = utils.load_admin_pw(CJDROUTE_CONF, CJDROUTE_BIN)
    c = cjdns.connect('127.0.0.1', 11234, admin_pw)

    result = c.nodeForAddr(addr)['result']

    try:
        x = cjdns.Peer(addr=addr)
    except ValueError:
        x = cjdns.Peer(ip=addr)

    for i in range(result['linkCount']):
        link = c.getLink(x.ip, i)

        if link and 'child' in link['result']:
            child = link['result']['child']
            if ip:
                y = cjdns.Peer(addr=child)
                yield '%s   %s  v%s' % (y.ip, y.path, y.version)
            else:
                yield child
        else:
            yield('-')

    c.disconnect()


cmd = [address, neighbours, ping, top, mon, route,
       sessions, search, uplinks]



from .arg import arg, wrap_errors
from . import xcjdns as cjdns
from . import cjdns as cj
from . import utils
from .const import YRD_INBOUND, YRD_OUTBOUND, CJDROUTE_CONF, CJDROUTE_BIN
import socket
import json
import sys
import os


@arg('name', help='name of the peer')
@arg('password', nargs='?', help='Set peering password')
@arg('-l', '--live', help='Don\'t write to disk')
@arg('-s', '--silent', help='ignore already added error')
@wrap_errors([socket.error, IOError])
def auth(name, password, live=False, silent=False):
    'add a password for inbound connections'

    if '/' in name:
        yield 'nope'
        exit(1)

    path = os.path.join(YRD_INBOUND, name)
    if os.path.exists(path):
        with open(path) as f:
            password = json.load(f)['password']
    else:
        if not password:
            password = utils.generate_key(31)

        info = {
            'name': name,
            'password': password
        }

        if not live:
            with open(path, 'w') as f:
                f.write(json.dumps(info))

    conf = utils.load_conf(CJDROUTE_CONF, CJDROUTE_BIN)
    c = cj.connect('127.0.0.1', 11234, conf['admin']['password'])
    resp = c.AuthorizedPasswords_add(user=name, password=password)
    try:
        utils.raise_on_error(resp)
    except:
        if not silent:
            raise
    c.disconnect()

    publicKey = conf['publicKey']
    port = conf['interfaces']['UDPInterface'][0]['bind'].split(':')[1]

    yield utils.to_credstr(utils.get_ip(), port, publicKey, password)


def resolve(source, path=None):
    if type(source) is list:
        return source

    source = source or path

    if not source:
        return []

    def read(source):
        source = sys.stdin if source == '-' else open(source)
        for auth in source:
            auth = auth.strip()

            if auth:
                yield auth

    return read(source)


@arg('name', help='the peers name')
@arg('source', nargs='?', help='read from file')
@arg('-l', '--live', help='Don\'t write to disk')
@wrap_errors([IOError])
def add(name, source, live=False):
    'add an outbound connection'
    if '/' in name:
        yield 'nope'
        exit(1)

    admin_pw = utils.load_admin_pw(CJDROUTE_CONF, CJDROUTE_BIN)
    c = cj.connect('127.0.0.1', 11234, admin_pw)

    path = os.path.join(YRD_OUTBOUND, name)
    out = not live and open(path, 'w')

    for auth in resolve(source, path):
        if out:
            out.write("%s\n" % auth)

        for addr, args in json.loads('{' + auth + '}').items():
            addr = utils.dns_resolve(addr)
            # TODO: this doesn't detect the proper interfaceNumber yet
            resp = c.UDPInterface_beginConnection(address=addr,
                                                  publicKey=args['publicKey'],
                                                  password=args['password'],
                                                  peerName=name,
                                                  interfaceNumber=0)
            utils.raise_on_error(resp)
    c.disconnect()


@wrap_errors([IOError])
def ls():
    'list passwords for inbound connections'
    admin_pw = utils.load_admin_pw(CJDROUTE_CONF, CJDROUTE_BIN)
    c = cjdns.connect('127.0.0.1', 11234, admin_pw)
    for user in c.listPasswords()['users']:
        yield user
    c.disconnect()


@wrap_errors([IOError])
def rm(user):
    'unpeer a node'
    if '/' in user:
        yield 'nope'
        exit(1)

    for path in [YRD_INBOUND, YRD_OUTBOUND]:
        path = os.path.join(path, user)
        if os.path.exists(path):
            os.unlink(path)
            yield 'deleted %r' % path

    admin_pw = utils.load_admin_pw(CJDROUTE_CONF, CJDROUTE_BIN)
    c = cj.connect('127.0.0.1', 11234, admin_pw)
    try:
        c.removePassword(user)
    except:
        pass
    c.disconnect()


cmd = [auth, add, ls, rm]



grey = '\x1b[30m'
red = '\x1b[31m'
green = '\x1b[32m'
yellow = '\x1b[33m'
reset = '\x1b[0m'

clear = '\x1b[2J\x1b[H'


class Session(object):
    def __init__(self):
        self.store = {}

    def update(self, line):
        addr = line[:39]
        path = line[40:59]

        try:
            return self.store[addr].update(path)
        except KeyError:
            self.store[addr] = Node(path)

    def fmt(self, line, signal):
        addr = line[:39]
        n = self.store[addr]
        return line[:40] + n.fmt(signal) + line[59:]

    def output(self, neighbours, lines, n):
        neighbours = [x[:39] + ' ' + Node(x[40:59]).fmt() + ' ' + x[60:64] for x in neighbours]

        buf = list(lines)
        buf.sort(key=lambda x: x[40:59], reverse=True)
        buf = [self.fmt(line, self.update(line)) for line in buf]

        output = clear

        if neighbours:
            output += '\n'.join(neighbours) + '\n\nNodestore (%d):\n' % len(buf)

        output += '\n'.join(buf[:n])

        return output


def cmp(a, b):
    return (a > b) - (a < b)


class Node(object):
    def __init__(self, path):
        self.path = path
        self.dist = self.calc(path)

    def update(self, path):
        if path != self.path:
            self.path = path
            o = self.dist
            self.dist = self.calc(path)
            return cmp(o, self.dist)

    def calc(self, path):
        return len(path.lstrip('0.'))

    def fmt(self, signal=None):
        cut = 19-self.dist
        c = reset if signal is None else green if signal > 0 else red if signal < 0 else yellow
        return grey + self.path[:cut] + c + self.path[cut:] + reset



from .arg import ArghParser, dispatch
from . import start
from . import core
from . import peer


parser = ArghParser(description='cjdns swiss army knife')
parser.add_commands(start.cmd + core.cmd)
parser.add_commands(peer.cmd, namespace='peer', title='ctrl peers')


def main():
    try:
        dispatch(parser)
    except KeyboardInterrupt as e:
        pass


if __name__ == '__main__':
    main()



import os

YRD_FOLDER = os.environ.get('YRD_FOLDER', '/etc/yrd')


def yrd(path):
    return os.path.join(YRD_FOLDER, path)


YRD_INBOUND = yrd('inbound.d/')
YRD_OUTBOUND = yrd('outbound.d/')

CJDROUTE_CONF = os.environ.get('CJDROUTE_CONF')
if not CJDROUTE_CONF:
    for d in [YRD_FOLDER, '/etc']:
        path = os.path.join(d, 'cjdroute.conf')
        try:
            os.stat(path)
        except OSError:
            pass
        else:
            break
    CJDROUTE_CONF = path
CJDROUTE_BIN = os.environ.get('CJDROUTE_BIN', 'cjdroute')



from .arg import arg, wrap_errors
from subprocess import Popen, PIPE
from .const import YRD_INBOUND, YRD_OUTBOUND, CJDROUTE_CONF, CJDROUTE_BIN
from . import utils
import json
import os


@arg('--attach', help='configure running cjdroute')
@wrap_errors([KeyboardInterrupt, IOError])
def start(attach=False, boot=False):
    'start and/or configure cjdroute'

    if not attach:
        p = Popen([CJDROUTE_BIN], stdin=PIPE)
        conf = utils.load_conf(CJDROUTE_CONF, CJDROUTE_BIN)
        p.communicate(json.dumps(conf).encode('utf-8'))

    from .peer import add
    for peer in os.listdir(YRD_OUTBOUND):
        yield '[*] connecting to %r' % peer
        list(add(peer, None, live=True))

    from .peer import auth
    for peer in os.listdir(YRD_INBOUND):
        yield '[*] adding %r' % peer
        list(auth(peer, None, silent=True, live=True))


cmd = [start]
